module BVNtools
  module Voyeur
    
    #require 'quick_magick.rb'
    
    if( not file_loaded?('makeTheFaces.rb') )
      # This will add a separator to the menu, but only once
      add_separator_to_menu('Voyeur')
      
      plugins_menu = UI.menu('Plugins')
      voyeur_menu = plugins_menu.add_submenu('Voyeur')
      voyeur_menu.add_item('make analysis faces'){ makeAnalysisFaces }
      voyeur_menu.add_item('Window Looker'){ start }
      voyeur_menu.add_item('Face Camera'){ placeUserCamera }
      
      toolbar = UI::Toolbar.new 'Voyeur'
      
      cmdFM = UI::Command.new('Face Maker'){ makeAnalysisFaces }
      cmdFM.small_icon = 'FM16.png'
      cmdFM.large_icon = 'FM24.png'
      cmdFM.tooltip = 'Makes offset faces for analysis'
      
      cmdWL = UI::Command.new('Window Looker'){ start }
      cmdWL.small_icon = 'WL16.png'
      cmdWL.large_icon = 'WL24.png'
      cmdWL.tooltip = 'photographs windows for their shadows'
            
      cmdUC = UI::Command.new('Face Camera'){ placeUserCamera }
      cmdUC.small_icon = 'camera16.gif'
      cmdUC.large_icon = 'camera24.gif'
      cmdUC.tooltip = 'points a camera at a face'
      
      toolbar = toolbar.add_item cmdFM
      toolbar = toolbar.add_item cmdWL
      toolbar = toolbar.add_item cmdUC
      toolbar.show
      
      file_loaded('makeTheFaces.rb')
    end

    class << self
      
      def isThisFourSided?(aFace)
        if aFace.vertices.length == 4
          return true
        else
          return false
        end
      end
      
      def getAnalysisDetails
        #this asks the user for some input to the process
        #OUTPUT it returns an array of strings that explain the various inputs
        #it'd be better/clearer if it returned a hash of formatted values
        #          [0                                        1           2                 3             4           5                 6                         7                   ]
        prompts  = ['image path',                                'cameraFOV','inset factor %', 'start hour', 'end hour', 'tests per hour', 'image width in pixels', 'Analysis Layer Name']
        defaults = ['C:\Users\bdoherty\Desktop\BVN\canberra flats\skTestImages','120',      '3',              '13',         '15',       '2',              '500',                   'Analysis'           ]
        lists    = ['',                                          '',         '',               '',           '',         '1|2|3|4|5',      '',                      ''                   ]
        input    = UI.inputbox prompts , defaults, lists, 'fill in some information'
        result = {:imagePath   => input[0].to_s.strip, 
                  :cameraFOV   => input[1].to_f,
                  :insetFactor => input[2].to_f * 0.01,
                  :startTime   => input[3].to_i,
                  :endTime     => input[4].to_i,
                  :hourDivs    => input[5].to_i,
                  :outputWidth => input[6].to_i,
                  :analysisLayerName => input[7].to_s
                 } 
        return result
      end
      
      def isFaceHorizontal?(aFace)
        #This function checks to see if a face is horizontal, 
        #i.e. if its normal faces directly up or directly down
        #INPUT it takes in a face, and an identifier that is  
        #used to provide an error if it is horizontal.
        #OUTPUT returns true if it is horizontal, false if it isn't
        internalNormal = aFace.normal
        if internalNormal == [0,0,1] || internalNormal == [0,0,-1]          
          return false
        else
          return true
        end
      end
    
      def calculateEyeDistance(theCamera, faceWidth)
        #calculates the distance that a camera needs to be away from 
        #a face in order to see it with a given field of view
        eyeDistance = ((faceWidth/2)/Math.tan(theCamera.fov.degrees/2))
        eyeDistance = eyeDistance/25.4 #to fix the crazy inch thing
        if eyeDistance == 0 
          eyeDistance = 10000          
        end
        return eyeDistance
      end
    
      def shadowInfoString
        #extract shadow information from the model to show to the user
        shadowInfo  = Sketchup.active_model.shadow_info
        message = ""#declare an empty string outside the scope of the 'each'
        shadowInfo.each_pair {|key, value| message += "#{key} is #{value}\n" }
        return message
      end
      
      def inspectImage(path, fileName, iar)
        myImage = QuickMagick::Image.read(path + fileName).first
        numColours = myImage.colors
        #i.convert "QuickMagick::Image.read(path+fileName).first -colorspace rgb -colors 10 -format \"%c\"  histogram:info:"
        #hold is there because it seems that if you assign to a variable, then the program waits 
        #for a return value, otherwise it just keeps going without anything to process.
        hold = myImage.draw_text(100, 100, 'colours ' << numColours.to_s)
        hold = myImage.save(path + fileName)
      end
      
      def findAMaterial(listOfAllMaterials, name)
      #INPUT   A string name of a material
      #RETURNS the material object that corresponds to the string name given
        for i in (0...listOfAllMaterials.length)
          if listOfAllMaterials[i].name == name
            return listOfAllMaterials[i]
          end
        end
      end
      
      def bound( valtoCheck, lowerBound, upperBound )
      #INPUT a number
      #RETURNS that number if it is between the boundaries, otherwise the boundary that it hits
      #i.e. bound(10,5,15) ==> 10
      #     bound(20,5,15) ==> 15
      #     bound( 0,5,15) ==>  5
        if valtoCheck > upperBound
          return upperBound
        elsif valtoCheck < lowerBound
          return lowerBound
        else
          return valtoCheck
        end
      end
    
    	def formatPath (pathString)
      #this adds a trailing / to the path if it doesn't have one
      #it also swaps the slashes from \ to /        
        pathString.gsub!("\\", '/')
        if pathString[pathString.length-1] == '/'
          puts '/ not added'
          return pathString
        else
          return pathString + '/'
        end
      end
  
      def getAnalysisDetailsFM
      #this asks the user for some input to the process
      #OUTPUT it returns an array of strings that explain the various inputs
      #            [ 0                       1            2              ]
        prompts  = ['Offset Distance (mm)', 'LayerName' , 'Material Name']
        defaults = ['10',                   'Analysis'  , 'Analysis'     ]
        lists    = ['',                     '',           ''             ]
        input    = UI.inputbox prompts , defaults, lists, "fill in some information"
        result = {
                  :offset   => input[0].to_f,
                  :layer    => input[1].to_s.strip,
                  :material => input[2].to_s.strip
                 } 
        return result
      end
      
      def findALayer(listOflayers, layerNameToMatch)
      #INPUT   A string name of a layer
      #RETURNS the layer object that corresponds to the string name given
        layerToMatch = nil
        listOflayers.each{|l|
          if l.name == layerNameToMatch 
            layerToMatch = l
          end
        }
        return layerToMatch
      end
      
      def layerFilter(setToFilter, layerNameToMatch)
        #INPUT   a set of entities to filter through & a string layername to search for
        #RETURNS an array of all the entities in the input set that are also on the input layer
        filteredSet = [] 
        #get the layer as an object so that the comparison 
        #isn't on the string name of the layer 
        layerToMatch = findALayer(Sketchup.active_model.layers, layerNameToMatch)
        #start filtering
        setToFilter.each{|e|
          if e.layer == layerToMatch
            filteredSet << e
          end
        }
        return filteredSet
      end
  
      def propertyFilter(setToFilter)
        newSet = []
        setToFilter.each{|e|
          if e.is_a? Sketchup::Face
            if isFaceHorizontal?(e) and isThisFourSided?(e)
              newSet << e
            end
          end
        }
      end

      def placeUserCamera
        #          [0       1              ]
        prompts  = ['FOV', 'inset factor %']
        defaults = ['120', '3'             ]
        lists    = ['',    ''              ]
        input    = UI.inputbox prompts , defaults, lists, 'tell me about this camera'
        result = {:cameraFOV   => input[1].to_f,
                  :insetFactor => input[2].to_f * 0.01}
        sel = Sketchup.active_model.selection
        if sel.length ==1
          if sel[0].is_a? Sketchup::Face
            camera = placeCamera(sel[0], result[:insetFactor], result[:cameraFOV],true)
            view = Sketchup.active_model.active_view
            status = view.camera = camera
          end
        else
          puts 'too many things selected\njust select one and try again'
        end
      end
      
      def placeCamera(face, insetFactor, cameraFOV, drawCamera)
        #get information about the face
        normal = face.normal
        centroidPoint = face.bounds.center
        #0 = left front bottom              #1 = right front bottom
        #2 = left back bottom               #3 = right back bottom
        #4 = left front top                 #5 = right front top
        #6 = left back top                  #7 = right back top
        height = (face.bounds.corner(4).distance face.bounds.corner(0)).to_mm
        width  = (face.bounds.corner(0).distance face.bounds.corner(3)).to_mm
        
        #setup for the camera
        camera = Sketchup::Camera.new              
        #camera.description  = 'camera looking at window ' << i.to_s
        aspectRatio = width/height
        camera.aspect_ratio = aspectRatio
        camera.fov = cameraFOV 
        #this shrinks the view to account for prudence and window frames etc.
        width = width * (1-insetFactor) 
        eyeDistance = calculateEyeDistance(camera, width)
        
        eyeOffset = normal.transform( Geom::Transformation.scaling(eyeDistance))
        
        eye          = centroidPoint.offset(eyeOffset)
        target       = centroidPoint
        unless drawCamera
          face.vertices.each{|v| line = Sketchup.active_model.entities.add_line eye,v }
        else
          camera.set eye, target, Z_AXIS
          return camera
        end
      end      
      

      def start
        puts "\n\n\n"
        
        result = UI.messagebox shadowInfoString << "\n\nAre these details correct?", MB_YESNO
        if result == 6 # Yes
          #this writes a message to the status bar (SB_PROMPT means the left bit)
          Sketchup.set_status_text "great, lets get going", SB_PROMPT
          
          userInput   = getAnalysisDetails                          #there needs to be / on the end of the path, formatPath does that
          imagePath   = formatPath(userInput[:imagePath])           #the field of view is in degrees and must be between 1 and 120)
          cameraFOV   = bound(userInput[:cameraFOV],  0, 120 )      #this trims the image ever so slightly so that it doesn't include the lines around the face
          insetFactor = bound(userInput[:insetFactor],0,50)         #this is percentages between 0 and 1, i.e. 50% is 0.5 and 5% is 0.05
          startTime   = bound(userInput[:startTime]  ,0,23)         #hours in 24hr format
          endTime     = bound(userInput[:endTime]    ,startTime,24) #hours in 24hr format
          hourDivs    =       userInput[:hourDivs]                  #how many chunks to chop the hour into i.e. 4 = 15 minutes
          outputWidth = bound(userInput[:outputWidth],50,2000)      #width in pixels of the images, height is set by the aspect ratio
          analysisLayerName = userInput[:analysisLayerName]         #the name of the layer that all the analysis faces are on
          #############################################################################################
          
          model = Sketchup.active_model
          #entities = model.active_entities
          #ent = model.entities
          filteredSelection = layerFilter(model.active_entities, analysisLayerName)
          filteredSelection = propertyFilter(filteredSelection)
          model.shadow_info["DisplayShadows"] = true
          m = 60/hourDivs
          
          itWorked = windowLooker(imagePath,         cameraFOV,   insetFactor,
                                  startTime,         endTime,     hourDivs, 
                                  filteredSelection, m,           model,
                                  outputWidth)
          if itWorked 
            puts "that all seemed to work out"
            Sketchup.set_status_text "that all seemed to work out", SB_PROMPT
          end
        else
          Sketchup.set_status_text "OHNOES!! Set the location settings and try again", SB_PROMPT
        end
      end #def start
    
      def windowLooker( imagePath,        cameraFOV,   insetFactor,
                        startTime,        endTime,     hourDivs, 
                        currentSelection, m,           model,
                        outputWidth)
                        
        Dir.chdir imagePath
        for i in (0...currentSelection.length)
          #builds a folder name from the current face's attributes
          folderName = imagePath + 'apptNum_'   + currentSelection[i].get_attribute("analysisInfo","apptNumber") +
                                   '_apptType_' + currentSelection[i].get_attribute("analysisInfo","apptType")
          #if the folder already exists, don't try and make it again!!
          if not File.directory? folderName
            Dir.mkdir(folderName)
            #puts 'made  ' + folderName
          end
          Dir.chdir folderName

          for hour in (startTime..endTime)
            for minutes in (0...hourDivs) # three dots ignores last value i.e. 0...3 ==> 0,1,2
              Sketchup.set_status_text "#{hour.to_s}:#{(m*minutes).to_s}", SB_PROMPT
              # make the face local for the rest of this loop
              face = currentSelection[i] 
              
              #set the time  Time.utc( year [, month, day, hour, min, sec, usec] )
              timeNow =      Time.utc( 2010,   "Jun", 21,  hour, m*minutes,  0)
              model.shadow_info["ShadowTime"] = timeNow 
              
              camera =  placeCamera(face, insetFactor, cameraFOV, true)
              
              #get the material and hold it, the face 
              #must be white to avoid problems of translucency
              holdMaterial = face.material              
              face.material = "snow"
              
              #change the view
              view = model.active_view
              status = view.camera = camera
              
              #save the image
              timeString = "#{"%02d" % hour}_#{("%02d" % (m*minutes))}"
              apptNum = 'apptNum_' + currentSelection[i].get_attribute("analysisInfo","apptNumber")
              fileName = apptNum + '_at_' + timeString + ".png" #'window' + ("%03d" %  i)
              view.write_image fileName, outputWidth, outputWidth*(1/camera.aspect_ratio), false
              
              #TODO this is all the image magick stuff, 
              #inspectImage(folderName, fileName, aspectRatio)
              
              #change the material back to what it was to begin with
              face.material = holdMaterial              
              view.refresh
              statusText = apptNum + " successful at #{timeString}" 
              puts statusText
              Sketchup.set_status_text statusText, SB_PROMPT
            end       #for minutes in (0...hourDivs)
          end         #for hour in (startTime..endTime)
        end           #for i in (0...currentSelection.length)
        Sketchup.active_model.entities.erase_entities(currentSelection.to_a)
      end             #windowLooker

      def makeAnalysisFaces
        
        mod = Sketchup.active_model
        ent = mod.entities
        sel = mod.selection
        
        userInput        = getAnalysisDetailsFM
        analysisLayer    = mod.layers.add userInput[:layer]
        analysisMaterial = findAMaterial(mod.materials, userInput[:material])
        faceCounter      = 0

        ent.each{ |e|
          if e.is_a? Sketchup::ComponentInstance
            e.definition.entities.each { |newE|
              if ((newE.is_a? Sketchup::Face) and (newE.material == analysisMaterial)) 
                #make an empty group
                tempGroup = ent.add_group
                #make an offset face
                normal = Geom::Vector3d.new(newE.normal).normalize
                offsetFactor = userInput[:offset].mm 
                offset = normal.transform( Geom::Transformation.scaling(offsetFactor))
                newPoints = []
                newE.vertices.each{|vertex| newPoints << vertex.position.offset(normal)}
                #make a new face in that group
                face = tempGroup.entities.add_face newPoints
                face.layer = analysisLayer
                #apply a transformation to the group taken from the instance of the component
                tempGroup.transformation = e.transformation
                #set some attributes for that face. They will be used later to indicate where the face came from.
                face.set_attribute "analysisInfo", "apptType",   e.definition.name
                face.set_attribute "analysisInfo", "apptNumber", e.name
                #remove the group, the face stays where it is
                tempGroup.explode 
                placeCamera(face, 0, 120, false)
                placeCamera(face, 0,  60, false)
                faceCounter = faceCounter + 1
              end       #if newE.is_a? Sketchup::Face
             }          #e.definition.entities.each
          end           #if e.is_a
        }               #ent.each
        status = "Made  #{faceCounter.to_s} planes"
        Sketchup.set_status_text status, SB_PROMPT
        puts status
      end               #makeAnalysisFaces
    end                 #class
  end                   #facemaker
end                     #voyeur
